@@ -56,7 +56,6 @@ import java.nio.charset.Charset;
 import java.security.Principal;
 import java.security.SecureRandom;
 import java.text.ParseException;
-import java.util.Collection;
 import java.util.Collections;
 import java.util.Enumeration;
 import java.util.HashMap;
@@ -67,22 +66,15 @@ import java.util.Locale;
 import java.util.Map;
 import java.util.Map.Entry;
 import java.util.Set;
-import javax.servlet.AsyncContext;
-import javax.servlet.DispatcherType;
 import javax.servlet.RequestDispatcher;
 import javax.servlet.ServletContext;
-import javax.servlet.ServletException;
 import javax.servlet.ServletInputStream;
-import javax.servlet.ServletRequest;
-import javax.servlet.ServletResponse;
 import javax.servlet.http.Cookie;
 import javax.servlet.http.HttpServlet;
 import javax.servlet.http.HttpServletRequest;
-import javax.servlet.http.HttpServletResponse;
 import javax.servlet.http.HttpSession;
-import javax.servlet.http.Part;
 import com.openexchange.ajp13.AJPv13Config;
-import com.openexchange.ajp13.AJPv13ServiceRegistry;
+import com.openexchange.ajp13.Services;
 import com.openexchange.ajp13.exception.AJPv13Exception;
 import com.openexchange.ajp13.exception.AJPv13Exception.AJPCode;
 import com.openexchange.ajp13.servlet.ServletConfigLoader;
@@ -93,6 +85,7 @@ import com.openexchange.config.ConfigTools;
 import com.openexchange.config.ConfigurationService;
 import com.openexchange.configuration.ServerConfig;
 import com.openexchange.configuration.ServerConfig.Property;
+import com.openexchange.dispatcher.Parameterizable;
 import com.openexchange.exception.OXException;
 import com.openexchange.java.DefaultHashKeyGenerator;
 import com.openexchange.java.HashKeyGenerator;
@@ -106,7 +99,7 @@ import com.openexchange.tools.servlet.http.Tools;
  *
  * @author <a href="mailto:thorben.betten@open-xchange.com">Thorben Betten</a>
  */
-public final class HttpServletRequestImpl implements HttpServletRequest {
+public final class HttpServletRequestImpl implements HttpServletRequest, Parameterizable {
 
     private static final org.apache.commons.logging.Log LOG =
         Log.valueOf(com.openexchange.log.LogFactory.getLog(HttpServletRequestImpl.class));
@@ -190,7 +183,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
 
     private HttpServlet servletInstance;
 
-    private long startTime;
+    private volatile long startTime;
 
     private ActionAwareServletInputStream servletInputStream;
 
@@ -215,6 +208,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
      */
     public HttpServletRequestImpl(final HttpServletResponseImpl response) {
         super();
+        contentLength = -1L;
         max = AJPv13Config.getMaxRequestParameterCount();
         contextPath = "";
         requestedSessionIdFromCookie = true;
@@ -228,10 +222,11 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         parameters = new HashKeyMap<List<String>>(max > 0 ? max : 64).setGenerator(hashKeyGenerator);
         headers = new HashMap<String, List<String>>(16);
         try {
-            setHeaderInternal(CONTENT_LENGTH, String.valueOf(-1), false);
+            setHeaderInternal(CONTENT_LENGTH, Integer.toString(-1), false);
         } catch (final AJPv13Exception e) {
             // Cannot occur
         }
+        startTime = System.currentTimeMillis();
     }
 
     /**
@@ -252,6 +247,15 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         servletInputStream.dumpToBuffer(bytes);
     }
 
+    /**
+     * Append specified bytes to buffer.
+     *
+     * @param bytes The bytes
+     */
+    public void appendToBuffer(final byte[] bytes) {
+        servletInputStream.appendToBuffer(bytes);
+    }
+
     /**
      * Checks if this request's content type indicates the form data: <code>"application/x-www-form-urlencoded"</code>
      *
@@ -274,7 +278,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         parameters.setGenerator(hashKeyGenerator);
         headers.clear();
         try {
-            setHeaderInternal(CONTENT_LENGTH, String.valueOf(-1), false);
+            setHeaderInternal(CONTENT_LENGTH, Integer.toString(-1), false);
         } catch (final AJPv13Exception e) {
             // Cannot occur
         }
@@ -307,7 +311,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         requestedSessionIdFromCookie = true;
         requestedSessionIdFromURL = false;
         servletInstance = null;
-        startTime = 0L;
+        startTime = System.currentTimeMillis();
         contentLength = -1L;
     }
 
@@ -318,7 +322,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
      */
     public void setContentLength(final long contentLength) {
         try {
-            setHeaderInternal(CONTENT_LENGTH, String.valueOf(contentLength), false);
+            setHeaderInternal(CONTENT_LENGTH, Long.toString(contentLength), false);
         } catch (final AJPv13Exception e) {
             // Cannot occur
         }
@@ -352,8 +356,20 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         }
     }
 
-    private static LinkedList<String> newLinkedList(final String initialValue) {
-        final LinkedList<String> list = new LinkedList<String>();
+    @Override
+    public void putParameter(String name, String value) {
+        if (null == name) {
+            throw new NullPointerException("name is null");
+        }
+        if (null == value) {
+            parameters.remove(name);
+        } else {
+            setParameter(name, value);
+        }
+    }
+
+    private static List<String> newLinkedList(final String initialValue) {
+        final List<String> list = new LinkedList<String>();
         list.add(initialValue);
         return list;
     }
@@ -446,8 +462,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
      */
     @Override
     public String getHeader(final String name) {
-        final String n = name.toLowerCase(Locale.ENGLISH);
-        return makeString(headers.get(n));
+        return makeString(headers.get(name.toLowerCase(Locale.ENGLISH)));
     }
 
     /**
@@ -467,7 +482,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
      * @return The header values as an {@link Enumeration}
      */
     @Override
-    public Enumeration<String> getHeaders(final String name) {
+    public Enumeration<?> getHeaders(final String name) {
         return makeEnumeration(headers.get(name.toLowerCase(Locale.ENGLISH)));
     }
 
@@ -477,7 +492,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
      * @return The header names as an {@link Enumeration}
      */
     @Override
-    public Enumeration<String> getHeaderNames() {
+    public Enumeration<?> getHeaderNames() {
         return makeEnumeration(headers.keySet().iterator());
     }
 
@@ -497,12 +512,12 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     }
 
     @Override
-    public Enumeration<String> getParameterNames() {
+    public Enumeration<?> getParameterNames() {
         return makeEnumeration(parameters.keySet().iterator());
     }
 
     @Override
-    public Map<String, String[]> getParameterMap() {
+    public Map<?, ?> getParameterMap() {
         final Map<String, String[]> retval = new HashKeyMap<String[]>(parameters.size()).setGenerator(hashKeyGenerator);
         for (final Entry<String, List<String>> entry : parameters.entrySet()) {
             final List<String> values = entry.getValue();
@@ -527,7 +542,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     }
 
     @Override
-    public Enumeration<String> getAttributeNames() {
+    public Enumeration<?> getAttributeNames() {
         return makeEnumeration(attributes.keySet().iterator());
     }
 
@@ -556,6 +571,10 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     @Override
     public void setCharacterEncoding(final String characterEncoding) throws UnsupportedEncodingException {
         String charset = characterEncoding;
+        if (null == charset) {
+            charset = "ISO-8859-1";
+            return;
+        }
         final int mlen;
         if (charset.charAt(0) == '"' && charset.charAt((mlen = charset.length() - 1)) == '"') {
             charset = charset.substring(1, mlen);
@@ -588,7 +607,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     }
 
     @Override
-    public Enumeration<Locale> getLocales() {
+    public Enumeration<?> getLocales() {
         return null;
     }
 
@@ -759,6 +778,20 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return values.get(0);
     }
 
+    private static final Enumeration EMPTY_ENUM = new Enumeration() {
+
+        @Override
+        public boolean hasMoreElements() {
+            return false;
+        }
+
+        @Override
+        public Object nextElement() {
+            return null;
+        }
+
+    };
+
     /**
      * Creates a new {@link Enumeration} for specified array.
      *
@@ -767,6 +800,9 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
      * @return A new {@link Enumeration}
      */
     protected static <T> Enumeration<T> makeEnumeration(final List<T> list) {
+        if (null == list) {
+            return EMPTY_ENUM;
+        }
         return (new Enumeration<T>() {
 
             private final int size = list.size();
@@ -896,7 +932,12 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     }
 
     public void setPathInfo(final String pathInfo) {
-        this.pathInfo = pathInfo;
+        if (null == pathInfo) {
+            this.pathInfo = null;
+        } else {
+            // Ensure starting slash '/' character
+            this.pathInfo = pathInfo.length() > 0 && '/' != pathInfo.charAt(0) ? '/' + pathInfo : pathInfo;
+        }
     }
 
     @Override
@@ -1047,8 +1088,8 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
          * First look-up HttpSessionManagement if a session already exists
          */
         final AjpProcessor ajpProcessor = response.getAjpProcessor();
-        final Cookie sessionCookie = ajpProcessor.getHttpSessionCookie();
-        final String httpSessionId = sessionCookie.getValue();
+        final Cookie jsessionIdCookie = ajpProcessor.getHttpSessionCookie();
+        final String httpSessionId = jsessionIdCookie.getValue();
         final HttpSessionWrapper httpSession = HttpSessionManagement.getHttpSession(httpSessionId);
         if (httpSession != null) {
             if (!HttpSessionManagement.isHttpSessionExpired(httpSession)) {
@@ -1069,38 +1110,64 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
             HttpSessionManagement.removeHttpSession(httpSessionId);
         }
         /*
-         * Create a new session
+         * Check whether to create a new session
          */
         if (create) {
-            /*
-             * Create new session
-             */
             session = ((HttpSessionWrapper) HttpSessionManagement.createAndGetHttpSession(httpSessionId));
             session.setNew(true);
             session.setServletContext(getServletContext());
             /*
              * Add JSESSIONID cookie
              */
-            configureCookie(sessionCookie);
-            response.addCookie(sessionCookie);
+            configureCookie(jsessionIdCookie);
+            response.addCookie(jsessionIdCookie);
         }
         return session;
     }
 
-    private static final String DEFAULT_PATH = "/";
+    private static volatile Boolean autologin;
 
-    private static void configureCookie(final Cookie sessionCookie) {
-        sessionCookie.setPath(DEFAULT_PATH);
-        final ConfigurationService configurationService = AJPv13ServiceRegistry.getInstance().getService(ConfigurationService.class);
+    private static boolean autologin() {
+        Boolean tmp = autologin;
+        if (null == tmp) {
+            synchronized (HttpServletResponseImpl.class) {
+                tmp = autologin;
+                if (null == tmp) {
+                    final ConfigurationService configurationService = Services.getService(ConfigurationService.class);
+                    tmp = Boolean.valueOf(null != configurationService && configurationService.getBoolProperty("com.openexchange.sessiond.autologin", false));
+                    autologin = tmp;
+                }
+            }
+        }
+        return tmp.booleanValue();
+    }
+
+    private static volatile Integer maxAge;
+
+    private static int maxAge() {
+        Integer tmp = maxAge;
+        if (null == tmp) {
+            synchronized (HttpServletResponseImpl.class) {
+                tmp = maxAge;
+                if (null == tmp) {
+                    final ConfigurationService service = Services.getService(ConfigurationService.class);
+                    tmp = Integer.valueOf(ConfigTools.parseTimespanSecs(null == service ?  "1W" : service.getProperty("com.openexchange.cookie.ttl", "1W")));
+                    maxAge = tmp;
+                }
+            }
+        }
+        return tmp.intValue();
+    }
+
+    public static void configureCookie(final Cookie jsessionIdCookie) {
         /*
          * Check if auto-login is enabled
          */
-        if (null != configurationService && configurationService.getBoolProperty("com.openexchange.sessiond.autologin", false)) {
-            final int maxAge =
-                (int) (ConfigTools.parseTimespan(configurationService.getProperty("com.openexchange.cookie.ttl", "1W")) / 1000);
-            sessionCookie.setMaxAge(maxAge);
+        if (autologin()) {
+            final int maxAge = maxAge();
+            jsessionIdCookie.setMaxAge(maxAge);
         } else {
-            sessionCookie.setMaxAge(-1); // cookies auto-expire
+            jsessionIdCookie.setMaxAge(-1); // cookies auto-expire
         }
     }
 
@@ -1154,8 +1221,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         }
     }
 
-    @Override
-    public ServletContext getServletContext() {
+    private final ServletContext getServletContext() {
         return ServletConfigLoader.getDefaultInstance().getContext(servletInstance.getClass().getCanonicalName(), servletPath);
     }
 
@@ -1177,59 +1243,4 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return startTime;
     }
 
-    @Override
-    public AsyncContext startAsync() throws IllegalStateException {
-        return null;
-    }
-
-    @Override
-    public AsyncContext startAsync(final ServletRequest servletRequest, final ServletResponse servletResponse) throws IllegalStateException {
-        return null;
-    }
-
-    @Override
-    public boolean isAsyncStarted() {
-        return false;
-    }
-
-    @Override
-    public boolean isAsyncSupported() {
-        return false;
-    }
-
-    @Override
-    public AsyncContext getAsyncContext() {
-        return null;
-    }
-
-    @Override
-    public DispatcherType getDispatcherType() {
-        return null;
-    }
-
-    @Override
-    public boolean authenticate(final HttpServletResponse response) throws IOException, ServletException {
-        return false;
-    }
-
-    @Override
-    public void login(final String username, final String password) throws ServletException {
-        // Empty
-    }
-
-    @Override
-    public void logout() throws ServletException {
-        // Empty
-    }
-
-    @Override
-    public Collection<Part> getParts() throws IOException, ServletException {
-        return Collections.emptyList();
-    }
-
-    @Override
-    public Part getPart(final String name) throws IOException, ServletException {
-        return null;
-    }
-
 }
